! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! May, 6th, 2014

! FD elastic operator module
! This is the main subroutine that calls
! the FD derivatives, sources, etc.
module elastic_mod

  use source_mod
  use seis_mod
  use elasticlib_mod
  use snap_mod
  use boundary_mod

  implicit none
  integer, private :: nt, nx, nz, sp, ti, abc
  integer, private :: xsource, zsource, nsource, nsnap, nseis
  real, private    :: dt, dx, dz, fp, source

  contains

  ! Initialization of all variables needed by this subroutine
  subroutine elastic_init(nz_in,dz_in,nx_in,dx_in,nt_in,dt_in,xsource_in,  &
                          zsource_in,nsource_in, fp_in, nsnap_in, nseis_in,&
                          sp_in, ti_in, abc_in)

    integer             :: nz_in,nx_in,nt_in,xsource_in,zsource_in,nsource_in
    integer             :: nsnap_in, nseis_in, sp_in, ti_in, abc_in
    real                :: dz_in,dx_in,dt_in, fp_in

    nz = nz_in
    dz = dz_in
    nt = nt_in
    dt = dt_in
    nx = nx_in
    dx = dx_in
    fp = fp_in
    xsource = xsource_in
    zsource = zsource_in
    nsource = nsource_in
    nsnap   = nsnap_in
    nseis = nseis_in
    sp = sp_in
    ti = ti_in
    abc = abc_in

  end subroutine

  ! Function that is called by Main
  function elastic_op(vp, vs, rho, seisp, seiss, snapvx, snapvz) result(stat)
    real,dimension(:,:)  :: vp, vs, rho, seisp, seiss
    real,dimension(:,:,:):: snapvx, snapvz
    integer              :: stat

    stat=1
    call elastic_op2(vp, vs, rho, seisp, seiss, snapvx, snapvz)
    stat=0
  end function

  ! Subroutine used by function
  subroutine elastic_op2(vp, vs, rho, seisp, seiss, snapvx, snapvz)
    integer                             :: it, ix, iz, dsnap, dseis
    integer                             :: countsnap,countseis, temp
    real, dimension(nz,nx)              :: vp,vs,rho
    real, dimension(nseis,nx-2*abc)     :: seisp, seiss
    real, dimension(nz,nx,nsnap)        :: snapvx,snapvz

    real, allocatable, dimension(:,:,:) :: mi, lambda
    real, allocatable, dimension(:,:,:) :: Vx,Vz,sigmaxx,sigmazz,sigmaxz
    real, allocatable, dimension(:,:,:) :: dxsigma,dzsigma,exx, ezz, exz
    real, allocatable, dimension(:,:,:) :: rhox, rhoz
    real                                :: aux1, t

    integer                             :: stat

    !Memory allocation

    !Allocate lame parameters
    allocate(mi(nz,nx,1))
    allocate(lambda(nz,nx,1))
    allocate(rhox(nz,nx,1))
    allocate(rhoz(nz,nx,1))

    !Allocate Velocities and stresses
    allocate(Vx(nz,nx,(ti-1)))
    allocate(Vz(nz,nx,(ti-1)))
    allocate(sigmaxx(nz,nx,(ti-1)))
    allocate(sigmazz(nz,nx,(ti-1)))
    allocate(sigmaxz(nz,nx,(ti-1)))

    !Allocate stress derivatives
    allocate(dxsigma(nz,nx,1))
    allocate(dzsigma(nz,nx,1))

    !Allocate strains
    allocate(exx(nz,nx,1))
    allocate(ezz(nz,nx,1))
    allocate(exz(nz,nx,1))

    !Initialize values
    mi     =0.0
    lambda =0.0
    Vx     =0.0
    Vz     =0.0
    sigmaxx=0.0
    sigmazz=0.0
    sigmaxz=0.0
    dxsigma=0.0
    dzsigma=0.0
    rhox   =0.0
    rhoz   =0.0
    exx    =0.0
    ezz    =0.0
    exz    =0.0
    dseis = (nt/nseis)+1
    dsnap = nt/nsnap
    countsnap = 1
    countseis = 1

    !Define Lame parameters
    aux1=(dt/dx)
    do ix=1,nx-1
      do iz=1,nz-1
        mi(iz,ix,1)=vs(iz,ix)*vs(iz,ix)*rho(iz,ix)
        lambda(iz,ix,1)=vp(iz,ix)*vp(iz,ix)*rho(iz,ix)-2.0*mi(iz,ix,1)
        rhox(iz,ix,1)=aux1*0.5*((1.0/rho(iz,ix+1)) + (1.0/rho(iz,ix)))
        rhoz(iz,ix,1)=aux1*0.5*((1.0/rho(iz+1,ix)) + (1.0/rho(iz,ix)))
      enddo
    enddo

    !Initialize subroutines
    call source_init(fp,dt,dx,nsource,nz,nx)
    call seis_init(nz,nx, dx, abc, seisp, seiss)
    call elasticmod_init(nz,nx,sp)
    call boundary_init(nz,nx,abc)
    call snap_init(nz,nx,snapvx,snapvz)
    
    !Time loop for 2nd order time
    write(0,*)"============================================================="
    write(0,*)"BEGIN OF TIME MODELING"
    write(0,*)"============================================================="
    do it=1,nt
      if (it<nsource) then
        call source_op2(it, source)
      else
        source=0.0
      endif

      !Particle velocities
      stat = dx_op(model=dxsigma,data=sigmaxx) !Due to simmetry, these
      stat = dz_op3(model=dzsigma,data=sigmaxz)!derivatives change, so
      Vx = Vx + rhox*(dxsigma+dzsigma)         !dx_op!=dx_op3
          
      stat = dx_op3(model=dxsigma,data=sigmaxz)
      stat = dz_op(model=dzsigma,data=sigmazz)
      Vz = Vz + rhoz*(dxsigma+dzsigma)

      !ntroduce absorbing boundary conditions
      stat = boundary_op(model=Vx,data=Vx)
      stat = boundary_op(model=Vz,data=Vz)

      !Strain components
      stat = dx_op3(model=exx,data=Vx)
      stat = dz_op3(model=ezz,data=Vz)
      stat = dxz_op(model=exz,data1=Vx, data2=Vz)

      !Stress components
      sigmaxx=sigmaxx+aux1*((lambda+2*mi)*exx+lambda*ezz)
      sigmazz=sigmazz+aux1*((lambda+2*mi)*ezz+lambda*exx)
      sigmaxz=sigmaxz+aux1*(mi*(exz))

      !Explosive source positioning
      source=100.0*source
      call explosive_source(source,xsource,zsource,sigmaxx,sigmazz)

      !Write seismograms
      if (mod(it,dseis)==0) then
        call seis_op2(countseis, Vx, Vz)
        countseis = countseis + 1
        write(0,*)"Seismogram time sample"
      endif

      !Write snapshots
      if (mod(it,dsnap)==0.and.countsnap.le.nsnap) then
        call snap_op2(countsnap,Vx,Vz)
        countsnap = countsnap + 1
      endif

      !Write to terminal
      write(0,*)"Time step is:",it," of ",nt
    enddo

    write(0,*)"Last seismogram sample is:",countseis

  end subroutine

end module
